/*
** Module   : QueryWrapper.java
** Abstract : Proxy for concrete query implementations
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20060430   @25885 Created initial version. Merged contents of
**                           AbstractQueryWrapper, CompoundQueryWrapper,
**                           PreselectQueryWrapper, and RandomAccessQuery-
**                           Wrapper. These classes are now obsolete. The
**                           merge was required because the same define
**                           query statement in Progress can be followed
**                           by many open query statements for that query;
**                           however, each open query may open a query of
**                           a different type than the previous one. Thus,
**                           a single wrapper type which supports all
**                           query types as delegates is required.
** 002 ECF 20060504   @25970 Added safety code. A default delegate query
**                           is automatically assigned upon construction,
**                           which has safe implementations of P2JQuery
**                           methods.
** 003 ECF 20060618   @27423 Change to assign() methods. Ensure query is
**                           set to scrolling upon assignment if the
**                           scrolling flag was previously set.
** 004 ECF 20061023   @30608 Added errorIfNull logic. P2JQuery interface
**                           now requires an implementation of the
**                           setErrorIfNull() method. The value set by the
**                           caller of this method is set into the query
**                           delegate, if one exists. If a delegate has
**                           not yet been assigned to the wrapper, the
**                           setting of this state in the delegate is
**                           deferred until a delegate is assigned.
** 005 ECF 20061027   @30754 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required
**                           for compile-time type safety.
** 006 ECF 20061102   @30960 Added support for reposition notification and
**                           auto-fetch features.
** 007 ECF 20061204   @31577 Implemented new P2JQuery methods to support
**                           browse.
** 008 NVS 20070502   @33345 Assigning a query should notify all listeners
**                           so that the related browse widget gets the
**                           notification and refreshes the display.
** 009 SIY 20070502   @33429 Added notifyRepositionListeners() method.
**                           Changed assignImpl() to use this new method.
**                           Minor cleanups in docs.
** 010 ECF 20070716   @34508 Added cleanup infrastructure. Wrapper is
**                           registered with TM at construction. At end of
**                           scope, cleanup is delegated to wrapped query.
**                           This ensures delegate does not clean up too
**                           early, since delegate may be assigned at a
**                           more deeply nested scope than those at which
**                           the wrapper can be used.
** 011 SIY 20070726   @34706 Helper method for explicit notification 
**                           invocations.
** 012 ECF 20080202   @37005 Changed constructor signature. Added argument
**                           to indicate whether delegate query is a
**                           scrolling query. This allows us to avoid an
**                           ugly, anonymous initializer in converted code
**                           which does nothing more than invoke the
**                           setScrolling() method. Also, automatically
**                           invoked delegate's open() method when a
**                           PreselectQuery delegate is assigned.
** 013 ECF 20080509   @38244 Fixed assignImpl(). Removed notification to
**                           reposition listeners. This is handled from
**                           the delegate's open() method, which is now
**                           part of P2JQuery and is called when the
**                           delegate is assigned.
** 014 ECF 20080509   @38253 Fixed regression introduced by #013 (@38244).
**                           assignImpl() must not call open() for a
**                           multi-table query, since this will trigger
**                           execution of the query too early. Deprecated
**                           notifyListeners(), which should be replaced
**                           in business logic with open().
** 015 ECF 20080512   @38495 Changed setScrolling(). Do not process method
**                           body more than once.
** 016 SVL 20080604   @38533 Modified according to the new P2JQuery
**                           interface.
** 017 ECF 20080729   @39327 Added rowid support. Refactored reposition*()
**                           methods.
** 018 ECF 20081106   @40325 Added _isOffEnd(). Required by P2JQuery
**                           interface change.
** 019 CA  20081210   @40840 Implemented CLOSE QUERY statement. On close,
**                           it must notify all browse widgets to reset 
**                           their state.
** 020 CA  20081211   @40860 Fixed regression introduced by H019: the 
**                           delegate needs to be explicitly closed, not
**                           cleaned.
** 021 CA  20090227   @41389 Added WrappedResources.isUnknown method; this
**                           implementation returns false.
** 022 SIY 20090306   @41425 Close delegate if new one is assigned. This is
**                           necessary to properly notify underlying browse
**                           when they are associated with query.
** 023 ECF 20090312   @42981 Implemented getTableCount(). Supports NUM-BUFFERS
**                           query handle attribute. Changed some integer
**                           parameters to recids. Use parameterized form of
**                           WeakList.
** 024 SVL 20090730   @43446 Removed open() call from assign(). It should be
**                           called externally.
** 025 CA  20090731   @43469 Register the query for off-end listener 
**                           registration after block initialization (done
**                           only for block types FOR and FOR EACH). Added
**                           getOffEndListeners, which returns a list of all
**                           off-end listeners for all used buffers.
** 026 SVL 20091030   @44289 Added getOffEnd() function to QueryWrapper and
**                           DefaultDelegate in order to support API change
**                           into P2JQuery.
** 027 SVL 20111111          presortDelegate checks that delegate is an
**                           instance of PresortDelegate (rather than
**                           PresortQuery).
** 028 VMN 20130207          Added support for handle-based methods:
**                           GET-BUFFER-HANDLE (KW_GET_BUFH)
**                           GET-CURRENT (KW_GET_CUR),
**                           QUERY-CLOSE (KW_QRY_CLOS),
**                           REPOSITION-FORWARD (KW_REPOS_F),
**                           REPOSITION-TO-ROW (KW_REPOS_2R)
**                           REPOSITION-TO-ROWID (KW_REPOS_2I),
**                           and attributes
**                           NUM-RESULTS (KW_NUM_RES)
**                           and attributes
**                           NUM-RESULTS (KW_NUM_RES)
** 029 OM  20130214          Added support for handle-based methods:
**                           QUERY-OPEN, QUERY-PREPARE, SET-BUFFERS, ADD-BUFFER
*/

package com.goldencode.p2j.persist;

import java.util.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.util.*;

/**
 * A delegating container implementation for {@link AbstractQuery}s.  An
 * instance of this class contains an instance of <code>AbstractQuery</code>,
 * to which it delegates the work of all of its methods.
 * <p>
 * The purpose of this class is to support queries defined in Progress using
 * the DEFINE QUERY statement.  These may be defined once, but opened and
 * closed multiple times.  In the P2J implementation, a new instance of
 * <code>AbstractQuery</code> is constructed each time a query is "opened"
 * (i.e., an OPEN QUERY was issued in the pre-conversion code).  Since a
 * reference to the query may be held by many routines, the query's internal
 * state must be updated when the query is opened or re-opened.  This is
 * accomplished by sharing a reference to an instance of this class, which
 * always holds the most recently constructed, underlying query.  Thus, since
 * all routines are accessing the query via this container (i.e., essentially
 * by proxy), they will always be accessing the correct, underlying query.
 * When a new query is opened, the new <code>AbstractQuery</code> instance is
 * stored in the wrapper via one of the <code>assign</code> method variants.
 * <p>
 * The public methods of this class may only be invoked if the underlying
 * delegate query is of a type which implements a method of the same
 * signature.  Many methods are supported by {@link AbstractQuery}, which is
 * the abstract base class of all concrete query implementations.  However,
 * certain methods are implemented only by particular implementations.  These
 * will fail with an <code>IllegalStateException</code> if invoked when the
 * wrong type of delegate has been set.
 * <p>
 * When an instance of this class is first created, a no-op, default
 * implementation of <code>P2JQuery</code> is assigned immediately.  This
 * ensures that at least those methods which are defined by the
 * <code>P2JQuery</code> interface have a safe methods.  However, this does
 * not protect the caller from invoking methods which are unique to specific
 * query types, if no safe delegate has first been assigned.  This safety
 * measure is intended to protect other runtime code which generically uses
 * an instance of this class via the <code>P2JQuery</code> interface (such as
 * <code>BrowseWidge</code> in the UI).
 * <p>
 * A query wrapped by this object may need to clean up resources when the
 * wrapper goes out of scope.  A cleaner object is registered with the
 * transaction manager upon construction of the wrapper, at the next external
 * scope.  This scope assumes that a <code>QueryWrapper</code> is only
 * instantiated as an instance member of a business logic class.  This is the
 * case for all converted Progress code.
 */
public final class QueryWrapper
implements P2JQuery
{
   /** Default delegate to use before a real one is assigned */
   private static final P2JQuery DEFAULT_DELEGATE = new DefaultDelegate();
   
   /** Delegate query */
   private P2JQuery delegate = null;
   
   /** Reposition listeners registered with the delegate query */
   private WeakList<RepositionListener> repoListeners =
      new WeakList<RepositionListener>();
   
   /** Should delegate be a scrolling query? */
   private boolean scrolling = false;
   
   /** Should delegate's failed first/last/unique retrievals raise error? */
   private boolean errorIfNull = true;
   
   /** Should the next row of the delegate be fetched after a reposition? */
   private boolean fetchOnReposition = false;
   
   /** Is query associated with a UI browse widget? */
   private boolean browsed = false;

   /** Flag indicating if the query is closing. */
   private boolean closing = false;

   /** Flag indicating if the query was explicitly closing. */
   private boolean closed = false;
   
   /**
    * Constructor which sets query scrolling mode for all delegate queries
    * assigned to this wrapper.
    * 
    * @param   scrolling
    *          <code>true</code> to set all delegate queries assigned to this
    *          wrapper to scrolling mode, else <code>false</code> to not use
    *          scrolling mode.
    */
   public QueryWrapper(boolean scrolling)
   {
      // Store a safe delegate until a real one is assigned.
      this.delegate = DEFAULT_DELEGATE;
      
      // Register a cleaner.  QueryWrappers are created as instance variables,
      // so we register to the next external scope, which corresponds to a
      // business logic class' execute() method.
      Finalizable cleaner = new Finalizable()
      {
         public void finished() { cleanup(); }
         public void iterate() {}
         public void retry() {}
      };
      TransactionManager.registerNextExternal(cleaner);
      
      if (scrolling)
      {
         setScrolling();
      }

      TransactionManager.registerOffEndQuery(this);
   }
   
   /**
    * Assign a newly opened <code>CompoundQuery</code> instance as this
    * container's active delegate query.
    *
    * @param   delegate
    *          New delegate query.
    */
   public void assign(CompoundQuery delegate)
   {
      assignImpl(delegate);
   }
   
   /**
    * Assign a newly opened <code>PreselectQuery</code> instance as this
    * container's active delegate query.
    *
    * @param   delegate
    *          New delegate query.
    */
   public void assign(PreselectQuery delegate)
   {
      assignImpl(delegate);
   }
   
   /**
    * Assign a newly opened <code>RandomAccessQuery</code> instance as this
    * container's active delegate query.
    *
    * @param   delegate
    *          New delegate query.
    */
   public void assign(RandomAccessQuery delegate)
   {
      assignImpl(delegate);
      delegate.setLenientOffEnd(true);
   }
   
   /**
    * Ensure that any <code>AbstractQuery</code> assigned to this wrapper is
    * a scrolling query, to enable results list scrolling/repositioning.  If
    * a query is already assigned to this wrapper, invoke its
    * <code>setScrolling</code> method immediately.
    */
   public void setScrolling()
   {
      if (!scrolling)
      {
         scrolling = true;
         if (delegate != null)
         {
            getDelegate().setScrolling();
         }
      }
   }
   
   /**
    * Open the prepared query delegate.
    */
   public void open()
   {
      delegate.open();
   }
   
   /**
    * Explicitly close the prepared query delegate.
    */
   public void close()
   {
      delegate.close();
      this.fetchOnReposition = true;
      this.closing = true;
      
      notifyRepositionListeners(true);
      
      this.delegate = DEFAULT_DELEGATE;
      this.closing = false;
      this.closed = true;
   }

   /**
    * Get the number of tables joined by this query.
    * 
    * @return  Number of tables joined by this query.
    */
   public int getTableCount()
   {
      return delegate.getTableCount();
   }
   
   /**
    * Get all the off-end listeners associated with this query.
    * 
    * @return  the off-end listeners associated with this query.
    */
   public List<QueryOffEndListener> getOffEndListeners()
   {
      return delegate.getOffEndListeners();
   }

   /**
    * Set a mode on the underlying buffers such that setting a null record
    * into a buffer may or may not result in an error condition.  This applies
    * only to <code>FIRST</code>, <code>LAST</code>, and <code>UNIQUE</code>
    * iteration types.
    * <p>
    * If a query delegate has already been set, it is updated with the new
    * setting immediately.  The setting is remembered, such that any delegate
    * assigned after this method is invoked will have this setting updated
    * upon assignment to the query wrapper.
    *
    * @param   errorIfNull
    *          If <code>true</code>, setting a <code>null</code> value as the
    *          current record in an underlying buffer will raise an error
    *          condition (if not in silent error mode); if <code>false</code>,
    *          this action will raise an end condition instead.
    */
   public void setErrorIfNull(boolean errorIfNull)
   {
      this.errorIfNull = errorIfNull;
      if (delegate != null)
      {
         getDelegate().setErrorIfNull(errorIfNull);
      }
   }
   
   /**
    * Add a joinable query to a compound query.  If this is not the first
    * component to be added to the query, its results will be joined to the
    * immediately previous component by an inner join.  This query will be
    * iterated using the <code>NEXT</code> iteration type.
    *
    * @param   query
    *          Query which will perform the fundamental work for this
    *          component when the compound query is iterated.
    *
    * @throws  IllegalStateException
    *          if this method is invoked during an iteration cycle or if
    *          delegate is not an instance of <code>CompoundQuery</code>.
    *
    * @see     #iterate
    */
   public void addComponent(Joinable query)
   {
      compoundDelegate().addComponent(query);
   }
   
   /**
    * Add a joinable query to a compound query, and specify the nature of
    * the iteration which will be performed (first, last, next, etc.) on that
    * component, when the compound query is iterated.  If this is not the
    * first component to be added to the query, its results will be joined to
    * the immediately previous component by an inner join.
    *
    * @param   query
    *          Query which will perform the fundamental work for this
    *          component when the compound query is iterated.
    * @param   iteration
    *          A {@link QueryConstants constant} indicating the type of work
    *          to be performed upon an iteration of the compound query.  This
    *          parameter determines whether the underlying query will be asked
    *          to perform a first, last, next, previous, or unique record
    *          retrieval.
    *
    * @throws  IllegalStateException
    *          if this method is invoked during an iteration cycle or if
    *          delegate is not an instance of <code>CompoundQuery</code>.
    *
    * @see     #iterate
    */
   public void addComponent(Joinable query, int iteration)
   {
      compoundDelegate().addComponent(query, iteration);
   }
   
   /**
    * Add a joinable query to a compound query, and specify the nature of
    * the iteration which will be performed (first, last, next, etc.) on that
    * component, when the compound query is iterated.  If this is not the
    * first component to be added to the query, the type of join performed
    * with the immediately previous component is determined by the value of
    * <code>outer</code>.
    *
    * @param   query
    *          Query which will perform the fundamental work for this
    *          component when the compound query is iterated.
    * @param   iteration
    *          A {@link QueryConstants constant} indicating the type of work
    *          to be performed upon an iteration of the compound query.  This
    *          parameter determines whether the underlying query will be asked
    *          to perform a first, last, next, previous, or unique record
    *          retrieval.
    * @param   outer
    *          <code>true</code> if this component should be joined to the
    *          previously added component via a left outer join;
    *          <code>false</code> if an inner join is required.  Note that
    *          adding an inner join component will cause any previously added
    *          components added as outer joins to be reset to inner joins.
    *
    * @throws  IllegalStateException
    *          if this method is invoked during an iteration cycle or if
    *          delegate is not an instance of <code>CompoundQuery</code>.
    *
    * @see     #iterate
    */
   public void addComponent(Joinable query, int iteration, boolean outer)
   {
      compoundDelegate().addComponent(query, iteration, outer);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo, String where)
   {
      preselectDelegate().addComponent(dmo, where);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse)
   {
      preselectDelegate().addComponent(dmo, where, inverse);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo, String where, Object[] args)
   {
      preselectDelegate().addComponent(dmo, where, args);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args)
   {
      preselectDelegate().addComponent(dmo, where, inverse, args);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            LockType lockType)
   {
      preselectDelegate().addComponent(dmo, where, lockType);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            LockType lockType)
   {
      preselectDelegate().addComponent(dmo, where, inverse, lockType);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            Object[] args,
                            LockType lockType)
   {
      preselectDelegate().addComponent(dmo, where, args, lockType);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType)
   {
      preselectDelegate().addComponent(dmo, where, inverse, args, lockType);
   }
   
   /**
    * Add a component to a preselect query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.  At
    *          this time, only <code>NEXT</code> is supported.  Future
    *          support for <code>FIRST</code>, <code>LAST</code>, and
    *          <code>UNIQUE</code> is expected.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      preselectDelegate().addComponent(dmo,
                                       where,
                                       args,
                                       lockType,
                                       iteration,
                                       outer);
   }
   
   /**
    * Add a component to a preselect query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.  At
    *          this time, only <code>NEXT</code> is supported.  Future
    *          support for <code>FIRST</code>, <code>LAST</code>, and
    *          <code>UNIQUE</code> is expected.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      preselectDelegate().addComponent(dmo,
                                       where,
                                       inverse,
                                       args,
                                       lockType,
                                       iteration,
                                       outer);
   }
   
   /**
    * Add a client-side where expression to a preselect query.  Each initial
    * query result found by the server must be accepted by this expression
    * before becoming a part of the query's final result set.
    *
    * @param   whereExpr
    *          Client-side where clause expression.  May not be
    *          <code>null</code>.
    *
    * @throws  NullPointerException
    *          if <code>whereExpr</code> is <code>null</code>.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addWhereExpression(WhereExpression whereExpr)
   {
      preselectDelegate().addWhereExpression(whereExpr);
   }
   
   /**
    * Set the iteration type for the most recently added query component to a
    * preselect query.
    *
    * @param   iteration
    *          Retrieval mode from {@link QueryConstants}.
    *
    * @throws  IllegalStateException
    *          If no component has yet been added, or
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void setIterationType(int iteration)
   {
      preselectDelegate().setIterationType(iteration);
   }
   
   /**
    * Add a single sort criterion, in the form of a <code>Resolvable</code>
    * object which is resolved at sort time.  It is the value of the resolved
    * result which is used for the sort.
    * <p>
    * The sort is ascending, from lowest resolved value to highest.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @param   sort
    *          A resolvable object whose resolved result is used for sorting.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public void addSortCriterion(Resolvable sort)
   {
      presortDelegate().addSortCriterion(sort);
   }
   
   /**
    * Add a single sort criterion, in the form of a <code>Resolvable</code>
    * object which is resolved at runtime.  It is the value of the resolved
    * result which is used for the sort.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @param   sort
    *          A resolvable object whose resolved result is used for sorting.
    * @param   descending
    *          if <code>true</code>, the sort is descending, from highest
    *          resolved value to lowest;  otherwise, the sort is ascending.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public void addSortCriterion(Resolvable sort, boolean descending)
   {
      presortDelegate().addSortCriterion(sort, descending);
   }
   
   /**
    * Enable all sort criteria set for this query to act as break group
    * categories.  When the query's results are presorted, the contents of
    * each break group are determined.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public void enableBreakGroups()
   {
      presortDelegate().enableBreakGroups();
   }
   
   /**
    * Add an accumulator to this query.  The accumulator expects to be
    * notified whenever a new result row is accessed (during sequential row
    * access).
    * <p>
    * This implementation assumes the accumulator is not associated with a
    * break group.
    *
    * @param   accum
    *          Accumulator to be registered with this query.
    */
   public void addAccumulator(Accumulator accum)
   {
      getDelegate().addAccumulator(accum);
   }
   
   /**
    * Add an accumulator to this query.  The accumulator expects to be
    * notified whenever a new result row is accessed (during sequential row
    * access).
    * <p>
    * This implementation assumes the accumulator is not associated with a
    * break group.
    *
    * @param   accum
    *          Accumulator to be registered with this query.
    * @param   deferred
    *          If <code>true</code>, calls to {@link #iterate()} will not
    *          cause the accumulator to recalculate immediately, but instead
    *          to update the accumulator's state such that the next invocation
    *          of {@link com.goldencode.p2j.util.Accumulator#accumulate()} will
    *          trigger the actual accumulation update.  If <code>false</code>,
    *          the accumulator recalculates its internal state immediately
    */
   public void addAccumulator(Accumulator accum, boolean deferred)
   {
      getDelegate().addAccumulator(accum, deferred);
   }
   
   /**
    * Reports if this query (that is, its delegate query) is valid for use.
    *
    * @return  <code>true</code> if this query's delegate is valid for use.
    */
   public boolean isValid()
   {
      return getDelegate().isValid();
   }
   
   /**
    * Reports if this object is unknown.<p>
    * This implementation always returns <code>false</code>.
    *
    * @return   <code>false</code>.
    */
   public boolean isUnknown()
   {
      return false;
   }

   /**
    * Advance a compound query, triggering one or more of the underlying
    * queries to retrieve a new record.  This produces a virtual, composite
    * row of results which is the combination of the records retrieved by
    * each component query.
    *
    * @throws  QueryOffEndException
    *          if no more records can be retrieved by the underlying queries.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void iterate()
   {
      compoundDelegate().iterate();
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters and
    * lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void first()
   {
      getDelegate().first();
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters, but
    * override the default lock type with that provided.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void first(LockType lockType)
   {
      getDelegate().first(lockType);
   }
   
   /**
    * Retrieve the first virtual, composite row of results for a compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void first(boolean iterating)
   {
      compoundDelegate().first(iterating);
   }
   
   /**
    * Retrieve the first virtual, composite row of results for a compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void first(boolean iterating, LockType lockType)
   {
      compoundDelegate().first(iterating, lockType);
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default lock type, but override the
    * default substitution parameters with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void first(Object[] values)
   {
      randomAccessDelegate().first(values);
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Override both the default substitution
    * parameters, and the default lock type with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void first(Object[] values, LockType lockType)
   {
      randomAccessDelegate().first(values, lockType);
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters and
    * lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void last()
   {
      getDelegate().last();
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters, but
    * override the default lock type with that provided.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void last(LockType lockType)
   {
      getDelegate().last(lockType);
   }
   
   /**
    * Retrieve the last virtual, composite row of results for a compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void last(boolean iterating)
   {
      compoundDelegate().last(iterating);
   }
   
   /**
    * Retrieve the last virtual, composite row of results for a compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void last(boolean iterating, LockType lockType)
   {
      compoundDelegate().last(iterating, lockType);
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default lock type, but override the
    * default substitution parameters with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void last(Object[] values)
   {
      randomAccessDelegate().last(values);
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Override both the default substitution
    * parameters, and the default lock type with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void last(Object[] values, LockType lockType)
   {
      randomAccessDelegate().last(values, lockType);
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters and
    * lock type.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    */
   public void next()
   {
      getDelegate().next();
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters, but
    * override the default lock type with that provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    */
   public void next(LockType lockType)
   {
      getDelegate().next(lockType);
   }
   
   /**
    * Retrieve the next virtual, composite row of results for a compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void next(boolean iterating)
   {
      compoundDelegate().next(iterating);
   }
   
   /**
    * Retrieve the next virtual, composite row of results for a compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void next(boolean iterating, LockType lockType)
   {
      compoundDelegate().next(iterating, lockType);
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default lock type, but override the
    * default substitution parameters with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void next(Object[] values)
   {
      randomAccessDelegate().next(values);
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Override both the default substitution
    * parameters, and the default lock type with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void next(Object[] values, LockType lockType)
   {
      randomAccessDelegate().next(values, lockType);
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Use the default substitution
    * parameters and lock type.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    */
   public void previous()
   {
      getDelegate().previous();
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Use the default substitution
    * parameters, but override the default lock type with that provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    */
   public void previous(LockType lockType)
   {
      getDelegate().previous(lockType);
   }
   
   /**
    * Retrieve the previous virtual, composite row of results for a compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void previous(boolean iterating)
   {
      compoundDelegate().previous(iterating);
   }
   
   /**
    * Retrieve the previous virtual, composite row of results for a compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void previous(boolean iterating, LockType lockType)
   {
      compoundDelegate().previous(iterating, lockType);
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Use the default lock type, but
    * override the default substitution parameters with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void previous(Object[] values)
   {
      randomAccessDelegate().previous(values);
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Override both the default
    * substitution parameters, and the default lock type with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void previous(Object[] values, LockType lockType)
   {
      randomAccessDelegate().previous(values, lockType);
   }
   
   /**
    * Reload the record most recently loaded into the backing record buffer,
    * if any.  Use the default lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record reload.
    */
   public void current()
   {
      getDelegate().current();
   }
   
   /**
    * Reload the record most recently loaded into the backing record buffer,
    * if any.  Override the default lock type.
    *
    * @param   lockType
    *          New lock type to apply to the record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record reload, or the lock could
    *          not be transitioned to the requested type.
    */
   public void current(LockType lockType)
   {
      getDelegate().current(lockType);
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Use the default substitution parameters and lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void unique()
   {
      getDelegate().unique();
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Use the default substitution parameters, but override
    * the default lock type with that provided.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void unique(LockType lockType)
   {
      getDelegate().unique(lockType);
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Use the default lock type, but override the default
    * substitution parameters with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void unique(Object[] values)
   {
      randomAccessDelegate().unique(values);
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Override both the default substitution parameters, and
    * the default lock type with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void unique(Object[] values, LockType lockType)
   {
      randomAccessDelegate().unique(values, lockType);
   }
   
   /**
    * Is this query set to fetch the next record after a reposition?
    *
    * @return  <code>true</code> if set to fetch after reposition, else
    *          <code>false</code>.
    */
   public boolean isFetchOnReposition()
   {
      return fetchOnReposition;
   }
   
   /**
    * Set this query to automatically fetch the next record after a
    * reposition, or turn this feature off.
    *
    * @param   enable
    *          <code>true</code> to enable automatic fetching;
    *          <code>false</code> to disable.
    */
   public void setFetchOnReposition(boolean enable)
   {
      fetchOnReposition = enable;
      if (delegate != null)
      {
         getDelegate().setFetchOnReposition(enable);
      }
   }
   
   /**
    * Register a listener to receive notifications reposition events for this
    * query.
    *
    * @param   listener
    *          Reposition listener to be registered.
    */
   public void addRepositionListener(RepositionListener listener)
   {
      repoListeners.add(listener);
      if (delegate != null)
      {
         getDelegate().addRepositionListener(listener);
      }
   }
   
   /**
    * Indicate to this query that whether it is associated with a browse
    * widget or not.
    *
    * @param   browsed
    *          <code>true</code> to indicate the query has an associated
    *          browse widget, else <code>false</code>.
    */
   public void setBrowsed(boolean browsed)
   {
      this.browsed = browsed;
      if (delegate != null)
      {
         getDelegate().setBrowsed(browsed);
      }
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified primary key ID.  A
    * request to retrieve the previous result will retrieve the result
    * immediately previous to this, if any, in the results list.
    *
    * @param   id
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * 
    * @throws  ErrorConditionException
    *          if the specified ID represents the unknown value.
    */
   public void repositionByID(recid id)
   {
      getDelegate().repositionByID(id);
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   id1
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * @param   joinIDs
    *          All remaining primary key IDs, if any, arranged from left to
    *          right to coincide with the records being joined by the
    *          underlying query.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if any of the specified IDs represents the unknown value.
    */
   public void repositionByID(rowid id1, rowid... joinIDs)
   {
      getDelegate().repositionByID(id1, joinIDs);
   }
   
   /**
    * Reposition the cursor to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the result at the specified row,
    * and a call to retrieve the previous result will get the result before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if <code>row</code> represents unknown value.
    */
   public void reposition(NumberType row)
   {
      getDelegate().reposition(row);
   }
   
   /**
    * Reposition the query to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the record after the new position,
    * and a call to retrieve the previous record will get the record before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public void reposition(int row)
   {
      getDelegate().reposition(row);
   }
   
   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if <code>rows</code> represents the unknown value.
    */
   public void forward(NumberType rows)
   {
      getDelegate().forward(rows);
   }
   
   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public void forward(int rows)
   {
      getDelegate().forward(rows);
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if <code>rows</code> represents the unknown value.
    */
   public void backward(NumberType rows)
   {
      getDelegate().backward(rows);
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public void backward(int rows)
   {
      getDelegate().backward(rows);
   }
   
   /**
    * Get the <b><i>current</i></b> size of the cached results list.  However,
    * since this cache is built progressively as the query is scrolled,
    * repositioned, and naturally navigated, this <b>does not</b> necessarily
    * represent the full count of results which may satisfy the query.
    *
    * @return  Current results list size or unknown value if the query is not
    *          open.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public integer size()
   {
      return getDelegate().size();
   }
   
   /**
    * Progress-compatible function that returns the 1-based index of the
    * current row in the result set. This is the index of the entry at or
    * before which the cursor currently is positioned.
    *
    * @return  Unknown value if the query is not open or result set is empty.
    *          1 if the query is positioned before the first record.
    *          A value 1 greater than the number of rows in the query result
    *          set if the query is positioned beyond the last record.
    *          Otherwise returns the number of the current row in the query
    *          result set.
    */
   public integer currentRow()
   {
      return getDelegate().currentRow();
   }
   
   /**
    * Get the 1-based index of the current row in the result set.  This is
    * the index of the entry at or before which the cursor currently is
    * positioned.
    *
    * @return  Current row or the unknown value if the current row cannot be
    *          determined.  The current row cannot be determined if the
    *          results list is empty, or if the cursor currently is in
    *          backward scroll mode.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public integer currentRowImpl()
   {
      return getDelegate().currentRowImpl();
   }
   
   /**
    * Indicate whether the cursor has run off one or the other end of the
    * associated query's list of available results.  If <code>true</code>,
    * this indicates that the query cannot produce any more results in the
    * current scroll direction.
    *
    * @return  <code>true</code> if the cursor is off either end of its
    *          query's results list.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public logical isOffEnd()
   {
      return getDelegate().isOffEnd();
   }
   
   /**
    * Indicate whether the cursor has run off one or the other end of the
    * associated query's list of available results.  If <code>true</code>,
    * this indicates that the query cannot produce any more results in the
    * current scroll direction.
    *
    * @return  <code>true</code> if the cursor is off either end of its
    *          query's results list.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public boolean _isOffEnd()
   {
      return getDelegate()._isOffEnd();
   }

   /**
    * Get the off-end status of the query, which indicates whether the query
    * has run off either end of its results.
    *
    * @return  One of the {@link OffEnd} constants <code>NONE</code>,
    *          <code>FRONT</code>, or <code>BACK</code>.
    */
   public OffEnd getOffEnd()
   {
      return getDelegate().getOffEnd();
   }

   /**
    * Determine whether the query result row currently being visited is the
    * first row in the presorted results list.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @return  <code>true</code> if the current result row is the first row,
    *          else <code>false</code>.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public logical isFirst()
   {
      return presortDelegate().isFirst();
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * first row within a break group whose category is identified by the
    * specified resolvable object.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @param   key
    *          Break group category key.  Must be the same object reference
    *          as was specified previously when invoking one of the
    *          <code>addSortCriterion</code> method variants.
    *
    * @return  <code>true</code> if current row is first in a break group;
    *          <code>false</code> if it is not;  unknown value if break groups
    *          have not been enabled.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public logical isFirstOfGroup(Resolvable key)
   {
      return presortDelegate().isFirstOfGroup(key);
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * last row in the presorted results list.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @return  <code>true</code> if the current result row is the last row,
    *          else <code>false</code>.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public logical isLast()
   {
      return presortDelegate().isLast();
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * first row within a break group whose category is identified by the
    * specified resolvable object.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @param   key
    *          Break group category key.  Must be the same object reference
    *          as was specified previously when invoking one of the
    *          <code>addSortCriterion</code> method variants.
    *
    * @return  <code>true</code> if current row is first in a break group;
    *          <code>false</code> if it is not;  unknown value if break groups
    *          have not been enabled.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public logical isLastOfGroup(Resolvable key)
   {
      return presortDelegate().isLastOfGroup(key);
   }
   
   /**
    * Indicate whether this query preselects all of its results.
    *
    * @return  <code>true</code> if the query preselects its results;
    *          <code>false</code> if the query's results are dynamically
    *          navigated.
    */
   public boolean isPreselect()
   {
      return getDelegate().isPreselect();
   }
   
   /**
    * Notify all registered listeners that query is repositioned or closed.
    * 
    * @param   closed
    *          A <code>true</code> value will inform all listeners that the
    *          query was explicitly closed.
    */
   public void notifyRepositionListeners(boolean closed)
   {
      delegate.notifyRepositionListeners(closed);
   }
   
   /**
    * Notify all registered listeners that query is repositioned in case if
    * query is attached to browse widget.
    * 
    * @deprecated Use #open() instead.
    */
   public void notifyListeners()
   {
      open();
   }
   
   /**
    * Indicate whether this query stands alone, or is the component of
    * another query, or is wrapped by a query wrapper.
    * 
    * @return  <code>true</code>;  query wrappers may not themselves be
    *          wrapped.
    */
   public boolean isStandalone()
   {
      return true;
   }
   
   /**
    * Change the standalone nature of this query.  A query is standalone by
    * default, but can be changed to non-standalone by the wrapper or
    * container which contains it.
    * <p>
    * Instances of <code>QueryWrapper</code> may not be wrapped in other
    * queries or query wrappers, so this implementation throws an exception
    * if this method is invoked with a parameter of <code>true</code>.
    * 
    * @param   standalone
    *          <code>true</code> to set query as standalone;
    *          <code>false</code> to set query as contained.
    *
    * @throws  UnsupportedOperationException
    *          if <code>standalone</code> is set to <code>true</code>.
    */
   public void setStandalone(boolean standalone)
   {
      if (!standalone)
      {
         throw new UnsupportedOperationException(
            "Query wrapper must be standalone");
      }
   }
   
   /**
    * Delegate cleanup processing to the query wrapped by this object.
    * 
    * @see  P2JQuery#cleanup()
    */
   public void cleanup()
   {
      if (!closed)
      {
         delegate.cleanup();
      }
   }

   /**
    * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
    *
    * Returns a handle of a particular buffer of a query object.
    *
    * @param   bufferName
    *          buffer-name. A CHARACTER expression that evaluates to the name
    *          of a buffer in the query or ProDataSet object.
    *
    * @return buffer handle
    */
   public handle bufferHandle(String bufferName)
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
    *
    * Returns a handle of a particular buffer of a query object.
    *
    * @param   bufferSequenceNumber
    *          buffer-sequence-number. An INTEGER that represents the sequence
    *          number of the desired buffer.
    *
    * @return buffer handle
    */
   public handle bufferHandle(int bufferSequenceNumber)
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds,
    *          else <code>false</code>.
    */
   public logical queryCurrent(int lock)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds,
    *          else <code>false</code>.
    */
   public logical queryCurrent(integer lock)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds,
    *          else <code>false</code>.
    */
   public logical queryCurrent(int lock, int nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds,
    *          else <code>false</code>.
    */
   public logical queryCurrent(integer lock, int nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds,
    *          else <code>false</code>.
    */
   public logical queryCurrent(integer lock, integer nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds,
    *          else <code>false</code>.
    */
   public logical queryCurrent(int lock, integer nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of QUERY-CLOSE() method (KW_QRY_CLOS).
    *
    * Closes a query object.
    *
    * @return  <code>true</code> if QUERY-CLOSE() succeeds,
    *          else <code>false</code>.
    */
   public logical queryClose()
   {
      return AbstractQuery.queryClose(this);
   }

   /**
    * Conversion of REPOSITION-FORWARD() method (KW_REPOS_F).
    *
    * Moves a query object’s result list pointer forward a particular number
    * of rows.
    *
    * @return  <code>true</code> if REPOSITION-FORWARD() succeeds,
    *          else <code>false</code>.
    */
   public logical queryForward(int rows)
   {
      return AbstractQuery.queryForward(this, rows);
   }

   /**
    * Conversion of REPOSITION-TO-ROW() method (KW_REPOS_2R).
    *
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified primary key ID.  A
    * request to retrieve the previous result will retrieve the result
    * immediately previous to this, if any, in the results list.
    *
    * @param   id
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    *
    * @return  <code>true</code> if REPOSITION-TO-ROW() succeeds,
    *          else <code>false</code>.
    */
   public logical queryReposition(rowid id)
   {
      return AbstractQuery.queryReposition(this, id);
   }

   /**
    * Conversion of REPOSITION-TO-ROWID() method (KW_REPOS_2I).
    *
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   id1
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * @param   joinIDs
    *          All remaining primary key IDs, if any, arranged from left to
    *          right to coincide with the records being joined by the
    *          underlying query.
    *
    * @return  <code>true</code> if REPOSITION-TO-ROWID() succeeds,
    *          else <code>false</code>.
    */
   public logical queryRepositionByID(rowid id1, rowid...joinIDs)
   {
      return AbstractQuery.queryRepositionByID(this, id1, joinIDs);
   }

   /**
    * Conversion of NUM-RESULTS attribute (KW_NUM_RES).
    *
    * Getter of NUM-RESULTS attribute
    *
    */
   public integer getNumResults()
   {
      // TODO:  implement
      return new integer();
   }

   /**
    * Adds one new buffer to a query object or dynamic ProDataSet object, 
    * without affecting the other buffers, if any. Use the SET-BUFFERS( )
    * method to remove all prior buffers and set all buffers for the object 
    * at the same time.
    * 
    * @param   buf
    *          A buffer to add.
    *
    * @return  true on success
    */
   @Override
   public logical addBuffer(Buffer buf)
   {
      // TODO Auto-generated method stub
      return new logical(false);
   }

   /**
    * Sets buffers for the query. Any buffers previously added
    * or set are removed. Use the <code>addBuffer()</code> method to add one 
    * buffer to the object, without affecting the other buffers, if any.
    * This method is the equivalent of <code>SET-BUFFERS()</code> method 
    * from 4GL.
    * 
    * @param   bufs
    *          Variable argument list of Buffer to be set.
    *          In 4GL, the maximum number of buffers per query is 18.
    * 
    * @return  true on success
    */
   @Override
   public logical setBuffers(handle... bufs)
   {
      // TODO Auto-generated method stub
      return new logical(false);
   }

   /**
    * Sets buffers for the query. Any buffers previously added
    * or set are removed. Use the <code>addBuffer()</code> method to add one 
    * buffer to the object, without affecting the other buffers, if any.
    * This method is the equivalent of <code>SET-BUFFERS()</code> method 
    * from 4GL.
    * 
    * @param   bufs
    *          Variable argument list of Buffer to be set.
    *          In 4GL, the maximum number of buffers per query is 18.
    * 
    * @return  true on success
    */
   @Override
   public logical setBuffers(Buffer... bufs)
   {
      // TODO Auto-generated method stub
      return new logical(false);
   }

   /**
    * Compiles a predicate (query condition). The query must be open after this call.
    * This method is the equivalent of <code>QUERY-PREPARE()</code> method from 4GL.
    *  
    * @param   predicate
    *          The predicate to be prepared by this query. The syntax of the string is the same
    *          of the <code>OPEN QUERY</code> statement of the 4GL.
    * 
    * @return  true on success
    */
   @Override
   public logical prepare(String predicate)
   {
      // TODO Auto-generated method stub
      return new logical(false);
   }

   /**
    * Compiles a predicate (query condition). The query must be open after this call.
    * This method is the equivalent of <code>QUERY-PREPARE()</code> method from 4GL.
    *  
    * @param   predicate
    *          The predicate to be prepared by this query. The syntax of the string is the same
    *          of the <code>OPEN QUERY</code> statement of the 4GL.
    * 
    * @return  true on success
    */
   @Override
   public logical prepare(character predicate)
   {
      // TODO Auto-generated method stub
      return new logical(false);
   }

   /**
    * Opens a query object. 
    * This method is the equivalent of <code>QUERY-OPEN()</code> method from 4GL.
    * <u>Note</u>: You must perform <code>prepare()</code> on a query 
    * object before you perform <code>queryOpen<code> on it. 
    * 
    * @return  true on success
    */
   @Override
   public logical queryOpen()
   {
      // TODO Auto-generated method stub
      return new logical(false);
   }

   /**
    * Assign a <code>P2JQuery</code> instance as this container's active
    * delegate query.  Set any deferred state which has been set for the
    * wrapper, such as scrollability and whether delegate should error out on
    * failed first/last/unique retrievals.
    *
    * @param   delegate
    *          New delegate query.
    */
   private void assignImpl(P2JQuery delegate)
   {
      if (this.closed == false && this.delegate != null)
      {
         close();
      }
      
      this.delegate = delegate;
      
      if (scrolling)
      {
         delegate.setScrolling();
      }
      
      delegate.setStandalone(false);
      delegate.setBrowsed(browsed);
      delegate.setErrorIfNull(errorIfNull);
      delegate.setFetchOnReposition(fetchOnReposition);
      
      Iterator<RepositionListener> iter = repoListeners.iterator();
      
      while (iter.hasNext())
      {
         RepositionListener next = iter.next();
         delegate.addRepositionListener(next);
      }
      
      this.closed = false;
   }

   /**
    * Convenience method to access the delegate query as an instance of
    * <code>CompoundQuery</code>.
    *
    * @return  Delegate query, cast to a <code>CompoundQuery</code>.
    *
    * @throws  IllegalStateException
    *          if the query delegate is not an instance of
    *          <code>CompoundQuery</code>.
    */
   private CompoundQuery compoundDelegate()
   {
      try
      {
         return (CompoundQuery) delegate;
      }
      catch (ClassCastException exc)
      {
         throw new IllegalStateException(
            "Expected compound query delegate;  found " +
            Utils.getUnqualifiedName(delegate.getClass()));
      }
   }
   
   /**
    * Convenience method to access the delegate query as an instance of
    * <code>PreselectQuery</code>.
    *
    * @return  Delegate query, cast to a <code>PreselectQuery</code>.
    *
    * @throws  IllegalStateException
    *          if the query delegate is not an instance of
    *          <code>PreselectQuery</code>.
    */
   private PreselectQuery preselectDelegate()
   {
      try
      {
         return (PreselectQuery) delegate;
      }
      catch (ClassCastException exc)
      {
         throw new IllegalStateException(
            "Expected preselect query delegate;  found " +
            Utils.getUnqualifiedName(delegate.getClass()));
      }
   }
   
   /**
    * Convenience method to access the delegate query as an instance of
    * <code>PresortDelegate</code>.
    *
    * @return  Delegate query, cast to a <code>PresortDelegate</code>.
    *
    * @throws  IllegalStateException
    *          if the query delegate is not an instance of
    *          <code>PresortDelegate</code>.
    */
   private PresortDelegate presortDelegate()
   {
      try
      {
         return (PresortDelegate) delegate;
      }
      catch (ClassCastException exc)
      {
         throw new IllegalStateException(
            "Expected presort query delegate;  found " +
            Utils.getUnqualifiedName(delegate.getClass()));
      }
   }
   
   /**
    * Convenience method to access the delegate query as an instance of
    * <code>RandomAccessQuery</code>.
    *
    * @return  Delegate query, cast to a <code>RandomAccessQuery</code>.
    *
    * @throws  IllegalStateException
    *          if the query delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   private RandomAccessQuery randomAccessDelegate()
   {
      try
      {
         return (RandomAccessQuery) delegate;
      }
      catch (ClassCastException exc)
      {
         throw new IllegalStateException(
            "Expected random access query delegate;  found " +
            Utils.getUnqualifiedName(delegate.getClass()));
      }
   }
   
   /**
    * Convenience method to access the current delegate query.
    *
    * @return  Delegate query assigned to this wrapper by business logic, or
    *          the default delegate query.
    */
   private P2JQuery getDelegate()
   {
      P2JQuery query = (closing ? DEFAULT_DELEGATE : delegate);
      
      return query;
   }
   
   /**
    * A no-op implementation of <code>P2JQuery</code> which is used as the
    * default delegate for the enclosing <code>QueryWrapper</code>.  This
    * implementation behaves as a query which returns no results would,
    * except that its <code>isValid</code> method returns <code>false</code>,
    * as an indication that it is a "fake" query for those callers to whom
    * this information is relevant.
    */
   private static class DefaultDelegate
   implements P2JQuery
   {
      /**
       * No-op.
       */
      public void setScrolling()
      {
      }
      
      /**
       * No-op.
       */
      public void open()
      {
      }
      
      /**
       * No-op.
       */
      public void close()
      {
      }
      
      /**
       * Get the number of tables joined by this query.
       * 
       * @return  0.
       */
      public int getTableCount()
      {
         return 0;
      }
      
      /**
       * Get all the off-end listeners associated with this query.
       * 
       * @return  the off-end listeners associated with this query.
       */
      public List<QueryOffEndListener> getOffEndListeners()
      {
         return Collections.EMPTY_LIST;
      }

     /**
       * No-op.
       *
       * @param   browsed
       *          Not used.
       */
      public void setBrowsed(boolean browsed)
      {
      }
      
      /**
       * No-op.
       *
       * @param   errorIfNull
       *          Not used.
       */
      public void setErrorIfNull(boolean errorIfNull)
      {
      }
      
      /**
       * No-op.
       *
       * @param   accum
       *          Not used.
       */
      public void addAccumulator(Accumulator accum)
      {
      }
      
      /**
       * No-op.
       *
       * @param   accum
       *          Not used.
       * @param   deferred
       *          Not used.
       */
      public void addAccumulator(Accumulator accum, boolean deferred)
      {
      }
      
      /**
       * No-op.
       */
      public void first()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void first(LockType lockType)
      {
      }
      
      /**
       * No-op.
       */
      public void last()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void last(LockType lockType)
      {
      }
      
      /**
       * No-op.
       */
      public void next()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void next(LockType lockType)
      {
      }
      
      /**
       * No-op.
       */
      public void previous()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void previous(LockType lockType)
      {
      }
      
      /**
       * No-op.
       */
      public void unique()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void unique(LockType lockType)
      {
      }
      
      /**
       * No-op.
       */
      public void current()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void current(LockType lockType)
      {
      }
      
      /**
       * Default implementation which reports query is not set to fetch the
       * next record after a reposition.
       *
       * @return  <code>false</code>.
       */
      public boolean isFetchOnReposition()
      {
         return false;
      }
      
      /**
       * No-op.
       *
       * @param   enable
       *          Not used.
       */
      public void setFetchOnReposition(boolean enable)
      {
      }
      
      /**
       * No-op.
       *
       * @param   listener
       *          Not used.
       */
      public void addRepositionListener(RepositionListener listener)
      {
      }
      
      /**
       * No-op.
       *
       * @param   id
       *          Not used.
       */
      public void repositionByID(recid id)
      {
      }
      
      /**
       * No-op.
       *
       * @param   id1
       *          Not used.
       * @param   joinIDs
       *          Not used.
       */
      public void repositionByID(rowid id1, rowid... joinIDs)
      {
      }
      
      /**
       * No-op.
       *
       * @param   row
       *          Not used.
       */
      public void reposition(NumberType row)
      {
      }
      
      /**
       * No-op.
       *
       * @param   row
       *          Not used.
       */
      public void reposition(int row)
      {
      }
      
      /**
       * No-op.
       *
       * @param   rows
       *          Not used.
       */
      public void forward(NumberType rows)
      {
      }
      
      /**
       * No-op.
       *
       * @param   rows
       *          Not used.
       */
      public void forward(int rows)
      {
      }
      
      /**
       * No-op.
       *
       * @param   rows
       *          Not used.
       */
      public void backward(NumberType rows)
      {
      }
      
      /**
       * No-op.
       *
       * @param   rows
       *          Not used.
       */
      public void backward(int rows)
      {
      }
      
      /**
       * Default implementation which reports unknown value to indicate a
       * closed query.
       *
       * @return  unknown value.
       */
      public integer size()
      {
         return (new integer());
      }
      
      /**
       * Default implementation which reports unknown value as the current
       * row to indicate an empty results list.
       *
       * @return  Unknown value.
       */
      public integer currentRow()
      {
         return currentRowImpl();
      }
      
      /**
       * Default implementation which reports unknown value as the current
       * row to indicate an empty results list.
       *
       * @return  Unknown value.
       */
      public integer currentRowImpl()
      {
         return (new integer());
      }
      
      /**
       * Default implementation which reports the query is off-end.
       *
       * @return  <code>true</code>.
       */
      public logical isOffEnd()
      {
         return (new logical(true));
      }
      
      /**
       * Default implementation which reports the query is off-end.
       *
       * @return  <code>true</code>.
       */
      public boolean _isOffEnd()
      {
         return true;
      }

      /**
       * Default implementation which reports that the query has the front
       * off-end.
       *
       * @return  {@link OffEnd#FRONT} constant.
       */
      public OffEnd getOffEnd()
      {
         return OffEnd.FRONT;
      }

      /**
       * Indicate whether this query preselects all of its results.
       *
       * @return  <code>false</code>.
       */
      public boolean isPreselect()
      {
         return false;
      }
      
      /**
       * Reports that this query is NOT valid for use.
       *
       * @return  <code>false</code>.
       */
      public boolean isValid()
      {
         return false;
      }
      
      /**
       * Reports if this object is unknown.  
       *
       * @return   <code>true</code> if object is unknown.
       */
      public boolean isUnknown()
      {
         return false;
      }

      /**
       * No-op.
       * 
       * @param   closed
       *          Not used.
       */
      public void notifyRepositionListeners(boolean closed)
      {
      }

      /**
       * Indicate whether this query stands alone, or is the component of
       * another query, or is wrapped by a query wrapper.
       * 
       * @return  <code>false</code>.
       */
      public boolean isStandalone()
      {
         return false;
      }
      
      /**
       * Change the standalone nature of this query.  A query is standalone by
       * default, but can be changed to non-standalone by the wrapper or
       * container which contains it.
       * <p>
       * This implementation is a no-op.
       * 
       * @param   standalone
       *          <code>true</code> to set query as standalone;
       *          <code>false</code> to set query as contained.
       */
      public void setStandalone(boolean standalone)
      {
      }
      
      /**
       * No-op.
       */
      public void cleanup()
      {
      }

      /**
       * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
       *
       * Returns a handle of a particular buffer of a query object.
       *
       * @param   bufferName
       *          buffer-name. A CHARACTER expression that evaluates to the name
       *          of a buffer in the query or ProDataSet object.
       *
       * @return  Unknown buffer handle.
       */
      public handle bufferHandle(String bufferName)
      {
         return (new handle());
      }

      /**
       * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
       *
       * Returns a handle of a particular buffer of a query object.
       *
       * @param   bufferSequenceNumber
       *          buffer-sequence-number. An INTEGER that represents the sequence
       *          number of the desired buffer.
       *
       * @return  Unknown buffer handle.
       */
      public handle bufferHandle(int bufferSequenceNumber)
      {
         return (new handle());
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       */
      public logical queryCurrent(int lock)
      {
         return (new logical(false));
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       */
      public logical queryCurrent(integer lock)
      {
         return (new logical(false));
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       */
      public logical queryCurrent(int lock, int nowait)
      {
         return (new logical(false));
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       */
      public logical queryCurrent(integer lock, int nowait)
      {
         return (new logical(false));
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       */
      public logical queryCurrent(integer lock, integer nowait)
      {
         return (new logical(false));
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       */
      public logical queryCurrent(int lock, integer nowait)
      {
         return (new logical(false));
      }

      /**
       * Conversion of QUERY-CLOSE() method (KW_QRY_CLOS).
       *
       * Closes a query object.
       *
       * @return  <code>false</code>.
       */
      public logical queryClose()
      {
         return (new logical(false));
      }

      /**
       * Conversion of REPOSITION-FORWARD() method (KW_REPOS_F).
       *
       * Moves a query object’s result list pointer forward a particular number
       * of rows.
       *
       * @return  <code>false</code>.
       */
      public logical queryForward(int rows)
      {
         return (new logical(false));
      }

      /**
       * Conversion of REPOSITION-TO-ROW() method (KW_REPOS_2R).
       *
       * Reposition the cursor such that a request to retrieve the next result
       * will retrieve the result which matches the specified primary key ID.  A
       * request to retrieve the previous result will retrieve the result
       * immediately previous to this, if any, in the results list.
       *
       * @param   id
       *          A primary key ID, representing the target record (in the case
       *          of a join, this ID is associated with the left-most record in
       *          the join).
       *
       * @return  <code>false</code>.
       */
      public logical queryReposition(rowid id)
      {
         return (new logical(false));
      }

      /**
       * Conversion of REPOSITION-TO-ROWID() method (KW_REPOS_2I).
       *
       * Reposition the cursor such that a request to retrieve the next result
       * will retrieve the result which matches the specified array of primary
       * key IDs.  A request to retrieve the previous result will retrieve the
       * result immediately previous to this, if any, in the results list.
       *
       * @param   id1
       *          A primary key ID, representing the target record (in the case
       *          of a join, this ID is associated with the left-most record in
       *          the join).
       * @param   joinIDs
       *          All remaining primary key IDs, if any, arranged from left to
       *          right to coincide with the records being joined by the
       *          underlying query.
       *
       * @return  <code>false</code>.
       */
      public logical queryRepositionByID(rowid id1, rowid...joinIDs)
      {
         return (new logical(false));
      }

      /**
       * Conversion of NUM-RESULTS attribute (KW_NUM_RES).
       *
       * Getter of NUM-RESULTS attribute
       *
       * @return  Unknown value.
       */
      public integer getNumResults()
      {
         return (new integer());
      }

      /**
       * Adds one new buffer to a query object or dynamic ProDataSet object, 
       * without affecting the other buffers, if any. Use the SET-BUFFERS( )
       * method to remove all prior buffers and set all buffers for the object 
       * at the same time.
       * 
       * @param   buf
       *          A buffer to add.
       *
       * @return  true on success
       */
      @Override
      public logical addBuffer(Buffer buf)
      {
         // TODO Auto-generated method stub
         return new logical(false);
      }

      /**
       * Sets buffers for the query. Any buffers previously added
       * or set are removed. Use the <code>addBuffer()</code> method to add one 
       * buffer to the object, without affecting the other buffers, if any.
       * This method is the equivalent of <code>SET-BUFFERS()</code> method 
       * from 4GL.
       * 
       * @param   bufs
       *          Variable argument list of Buffer to be set.
       *          In 4GL, the maximum number of buffers per query is 18.
       * 
       * @return  true on success
       */
      @Override
      public logical setBuffers(handle... bufs)
      {
         // TODO Auto-generated method stub
         return new logical(false);
      }

      /**
       * Sets buffers for the query. Any buffers previously added
       * or set are removed. Use the <code>addBuffer()</code> method to add one 
       * buffer to the object, without affecting the other buffers, if any.
       * This method is the equivalent of <code>SET-BUFFERS()</code> method 
       * from 4GL.
       * 
       * @param   bufs
       *          Variable argument list of Buffer to be set.
       *          In 4GL, the maximum number of buffers per query is 18.
       * 
       * @return  true on success
       */
      @Override
      public logical setBuffers(Buffer... bufs)
      {
         // TODO Auto-generated method stub
         return new logical(false);
      }

      /**
       * Compiles a predicate (query condition). This method is
       * the equivalent of <code>QUERY-PREPARE()</code> method from 4GL.
       * 
       * @param   predicate
       *          The predicate to be prepared by this query.
       * 
       * @return  true on success
       */
      @Override
      public logical prepare(String predicate)
      {
         // TODO Auto-generated method stub
         return new logical(false);
      }

      /**
       * Compiles a predicate (query condition). This method is
       * the equivalent of <code>QUERY-PREPARE()</code> method from 4GL.
       * 
       * @param   predicate
       *          The predicate to be prepared by this query.
       * 
       * @return  true on success
       */
      @Override
      public logical prepare(character predicate)
      {
         // TODO Auto-generated method stub
         return new logical(false);
      }

      /**
       * Opens a query object. 
       * This method is the equivalent of <code>QUERY-OPEN()</code> method from 4GL.
       * <u>Note</u>: You must perform <code>prepare()</code> on a query 
       * object before you perform <code>queryOpen<code> on it. 
       * 
       * @return  true on success
       */
      @Override
      public logical queryOpen()
      {
         // TODO Auto-generated method stub
         return new logical(false);
      }
   }
}
